from __future__ import annotations

import functools
from bs4.element import _PageElementT, _Strainable, SoupStrainer, NavigableString, ResultSet
from abc import ABCMeta, abstractmethod

from _typeshed import Incomplete, ReadableBuffer
from collections.abc import Callable, Iterable, Iterator
from re import Pattern
from typing import Any, Generic, TypeVar, overload, List
from typing_extensions import Self, SupportsIndex

from bs4 import BeautifulSoup
from bs4.builder import TreeBuilder
from bs4.formatter import Formatter, _EntitySubstitution
from bs4.element import PageElement, Tag

from .accept_everything import FullDunder

DEFAULT_OUTPUT_ENCODING: str
nonwhitespace_re: Pattern[str]
whitespace_re: Pattern[str]
PYTHON_SPECIFIC_ENCODINGS: set[str]

ChainingBroadcastListType = list
T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)


class BroadcastList(list[T]):
    @property
    def bc(self) -> _BroadcastedList[T]:
        ...


class _BroadcastedList(FullDunder, Generic[T_co]):
    def _callable_attr_broadcast(self, *args, **kwargs) -> BroadcastList:
        ...

    def __init__(self, broadcastlist: BroadcastList[T_co]) -> None:
        ...

    def __getattr__(self, __name: str) -> Callable[..., BroadcastList] | BroadcastList:
        ...

    def _callable_dunder_getattr(self, __name: str, *args, **kwargs) -> Any:
        ...

    async def _callable_dunder_getattr_async(self, __name: str, *args, **kwargs) -> Any:
        ...

    def __setattr__(self, name: str, value) -> None:
        ...

    def __str__(self) -> str:
        ...


class NewTagBroadcastList(list[Tag]):
    @property
    def bc(self) -> _TagBroadcastedList:
        ...


class _TagBroadcastedList(_BroadcastedList[Tag]):
    """Chaining BroadcastED list especially for Tags."""
    parent: BroadcastList[Tag | None]
    previous_element: BroadcastList[PageElement | None]
    next_element: BroadcastList[PageElement | None]
    next_sibling: BroadcastList[PageElement | None]
    previous_sibling: BroadcastList[PageElement | None]
    def setup(
        self,
        parent: Tag | None = None,
        previous_element: PageElement | None = None,
        next_element: PageElement | None = None,
        previous_sibling: PageElement | None = None,
        next_sibling: PageElement | None = None,
    ) -> BroadcastList[None]: ...
    def format_string(self, s: str, formatter: Formatter | str | None) -> BroadcastList[str]: ...
    def formatter_for_name(self, formatter: Formatter | str | _EntitySubstitution) -> BroadcastList[None]: ...
    nextSibling: BroadcastList[PageElement | None]
    previousSibling: BroadcastList[PageElement | None]
    @property
    def stripped_strings(self) -> BroadcastList[Iterator[str]]: ...
    def get_text(self, separator: str = "", strip: bool = False, types: tuple[type[NavigableString], ...] = ...) -> BroadcastList[str]: ...
    getText = get_text
    @property
    def text(self) -> BroadcastList[str]: ...
    def replace_with(self, *args: PageElement | str) -> BroadcastList[Self]: ...
    replaceWith = replace_with
    def unwrap(self) -> BroadcastList[Self]: ...
    replace_with_children = unwrap
    replaceWithChildren = unwrap
    def wrap(self, wrap_inside: _PageElementT) -> BroadcastList[_PageElementT]: ...
    def extract(self, _self_index: int | None = None) -> BroadcastList[Self]: ...
    def Einsert(self, position: int, new_child: PageElement | str) -> BroadcastList[None]: ...
    def Eappend(self, tag: PageElement | str) -> BroadcastList[None]: ...
    def Eextend(self, tags: Iterable[PageElement | str]) -> BroadcastList[None]: ...
    def insert_before(self, *args: PageElement | str) -> BroadcastList[None]: ...
    def insert_after(self, *args: PageElement | str) -> BroadcastList[None]: ...
    def find_next(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | NavigableString | None]: ...
    findNext = find_next
    def find_all_next(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[PageElement]]: ...
    findAllNext = find_all_next
    def find_next_sibling(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | NavigableString | None]: ...
    findNextSibling = find_next_sibling
    def find_next_siblings(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[PageElement]]: ...
    findNextSiblings = find_next_siblings
    fetchNextSiblings = find_next_siblings
    def find_previous(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | NavigableString | None]: ...
    findPrevious = find_previous
    def find_all_previous(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[PageElement]]: ...
    findAllPrevious = find_all_previous
    fetchPrevious = find_all_previous
    def find_previous_sibling(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | NavigableString | None]: ...
    findPreviousSibling = find_previous_sibling
    def find_previous_siblings(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[PageElement]]: ...
    findPreviousSiblings = find_previous_siblings
    fetchPreviousSiblings = find_previous_siblings
    def find_parent(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | None]: ...
    findParent = find_parent
    def find_parents(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[Tag]]: ...
    findParents = find_parents
    fetchParents = find_parents
    @property
    def next(self) -> BroadcastList[Tag | NavigableString | None]: ...
    @property
    def previous(self) -> BroadcastList[Tag | NavigableString | None]: ...
    @property
    def next_elements(self) -> BroadcastList[Iterable[PageElement]]: ...
    @property
    def next_siblings(self) -> BroadcastList[Iterable[PageElement]]: ...
    @property
    def previous_elements(self) -> BroadcastList[Iterable[PageElement]]: ...
    @property
    def previous_siblings(self) -> BroadcastList[Iterable[PageElement]]: ...
    @property
    def parents(self) -> BroadcastList[Iterable[Tag]]: ...
    @property
    def decomposed(self) -> BroadcastList[bool]: ...
    def nextGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def nextSiblingGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def previousGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def previousSiblingGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def parentGenerator(self) -> BroadcastList[Iterable[Tag]]: ...

    # TAGS

    parser_class: BroadcastList[type[BeautifulSoup] | None]
    name: BroadcastList[str]
    namespace: BroadcastList[str | None]
    prefix: BroadcastList[str | None]
    sourceline: BroadcastList[int | None]
    sourcepos: BroadcastList[int | None]
    known_xml: BroadcastList[bool | None]
    attrs: BroadcastList[dict[str, str]]
    contents: BroadcastList[list[PageElement]]
    hidden: BroadcastList[bool]
    can_be_empty_element: BroadcastList[bool | None]
    cdata_list_attributes: BroadcastList[list[str] | None]
    preserve_whitespace_tags: BroadcastList[list[str] | None]
    # def __init__(
    #     self,
    #     parser: BeautifulSoup | None = None,
    #     builder: TreeBuilder | None = None,
    #     name: str | None = None,
    #     namespace: str | None = None,
    #     prefix: str | None = None,
    #     attrs: dict[str, str] | None = None,
    #     parent: Tag | None = None,
    #     previous: PageElement | None = None,
    #     is_xml: bool | None = None,
    #     sourceline: int | None = None,
    #     sourcepos: int | None = None,
    #     can_be_empty_element: bool | None = None,
    #     cdata_list_attributes: list[str] | None = None,
    #     preserve_whitespace_tags: list[str] | None = None,
    #     interesting_string_types: type[NavigableString] | tuple[type[NavigableString], ...] | None = None,
    #     namespaces: dict[str, str] | None = None,
    # ) -> NewTagBroadcastList[None]: ...
    parserClass: BroadcastList[type[BeautifulSoup] | None]
    def __copy__(self) -> BroadcastList[Self]: ...
    @property
    def is_empty_element(self) -> BroadcastList[bool]: ...
    @property
    def isSelfClosing(self) -> BroadcastList[bool]: ...
    @property
    def string(self) -> BroadcastList[str | None]: ...
    @string.setter
    def string(self, string: str) -> BroadcastList[None]: ...
    DEFAULT_INTERESTING_STRING_TYPES: BroadcastList[tuple[type[NavigableString], ...]]
    @property
    def strings(self) -> BroadcastList[Iterable[str]]: ...
    def decompose(self) -> BroadcastList[None]: ...
    def Eclear(self, decompose: bool = False) -> BroadcastList[None]: ...
    def smooth(self) -> BroadcastList[None]: ...
    def Eindex(self, element: PageElement) -> BroadcastList[int]: ...
    def get(self, key: str, default: str | list[str] | None = None) -> BroadcastList[str | list[str] | None]: ...
    def get_attribute_list(self, key: str, default: str | list[str] | None = None) -> BroadcastList[list[str]]: ...
    def has_attr(self, key: str) -> BroadcastList[bool]: ...
    def __hash__(self) -> BroadcastList[int]: ...
    def __getitem__(self, key: str) -> BroadcastList[str | list[str]]: ...
    def __iter__(self) -> BroadcastList[Iterator[PageElement]]: ...
    def __len__(self) -> BroadcastList[int]: ...
    def __contains__(self, x: object) -> BroadcastList[bool]: ...
    def __bool__(self) -> BroadcastList[bool]: ...
    def __setitem__(self, key: str, value: str | list[str]) -> BroadcastList[None]: ...
    def __delitem__(self, key: str) -> BroadcastList[None]: ...
    def __getattr__(self, tag: str) -> BroadcastList[Tag | None]: ...
    def __eq__(self, other: object) -> BroadcastList[bool]: ...
    def __ne__(self, other: object) -> BroadcastList[bool]: ...
    def __unicode__(self) -> BroadcastList[str]: ...
    def encode(
        self,
        encoding: str = "utf-8",
        indent_level: int | None = None,
        formatter: str | Formatter = "minimal",
        errors: str = "xmlcharrefreplace",
    ) -> BroadcastList[bytes]: ...
    def decode(
        self,
        indent_level: int | None = None,
        eventual_encoding: str = "utf-8",
        formatter: str | Formatter = "minimal",
        iterator: Iterator[PageElement] | None = None,
    ) -> BroadcastList[str]: ...
    @overload
    def prettify(self, encoding: str, formatter: str | Formatter = "minimal") -> BroadcastList[bytes]: ...
    @overload
    def prettify(self, encoding: None = None, formatter: str | Formatter = "minimal") -> BroadcastList[str]: ...
    def decode_contents(
        self, indent_level: int | None = None, eventual_encoding: str = "utf-8", formatter: str | Formatter = "minimal"
    ) -> BroadcastList[str]: ...
    def encode_contents(
        self, indent_level: int | None = None, encoding: str = "utf-8", formatter: str | Formatter = "minimal"
    ) -> BroadcastList[bytes]: ...
    def renderContents(self, encoding: str = "utf-8", prettyPrint: bool = False, indentLevel: int = 0) -> BroadcastList[bytes]: ...
    def find(
        self,
        name: _Strainable | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        recursive: bool = True,
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[Tag | NavigableString | None]: ...
    findChild = find
    def find_all(
        self,
        name: _Strainable | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        recursive: bool = True,
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> BroadcastList[ResultSet[Any]]: ...
    __call__ = find_all
    findAll = find_all
    findChildren = find_all
    @property
    def children(self) -> BroadcastList[Iterable[PageElement]]: ...
    @property
    def descendants(self) -> BroadcastList[Iterable[PageElement]]: ...
    def select_one(
        self, selector: str, namespaces: Incomplete | None = None, *, flags: int = ..., custom: dict[str, str] | None = ...
    ) -> BroadcastList[Tag | None]: ...
    def select(
        self,
        selector: str,
        namespaces: Incomplete | None = None,
        limit: int | None = None,
        *,
        flags: int = ...,
        custom: dict[str, str] | None = ...,
    ) -> BroadcastList[ResultSet[Tag]]: ...
    def childGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def recursiveChildGenerator(self) -> BroadcastList[Iterable[PageElement]]: ...
    def has_key(self, key: str) -> BroadcastList[bool]: ...

#########################
# LEGACY BROADCAST LIST #
#########################

class AbstractBroadcastList(list[T], metaclass=ABCMeta):
    @abstractmethod
    def _callable_attr_broadcast(self, *args, attr_name: str = '', **kwargs) -> Any:
        ...

    @abstractmethod
    def _attr_broadcast(self, attr_name: str) -> Any:
        ...

    def __getattr__(self, __name: str):
        ...

class NonchainingBroadcastList(AbstractBroadcastList[T]):
    def _callable_attr_broadcast(self, *args, attr_name: str = '', **kwargs) -> None:
        ...

    def _attr_broadcast(self, attr_name: str) -> None:
        ...

    @overload
    def __getitem__(self, __item: SupportsIndex) -> T: ...
    @overload
    def __getitem__(self, __item: slice) -> list[T]: ...
    @overload
    def __getitem__(self, __item) -> list: ...

class ChainingBroadcastList(AbstractBroadcastList[T]):
    def _callable_attr_broadcast(self, *args, attr_name: str = '', **kwargs) -> ChainingBroadcastList:
        ...

    def _attr_broadcast(self, attr_name: str) -> ChainingBroadcastList:
        ...

    @overload
    def __getitem__(self, __item: SupportsIndex) -> T: ...
    @overload
    def __getitem__(self, __item: slice) -> Self: ...
    @overload
    def __getitem__(self, __item) -> ChainingBroadcastList: ...

class TagBroadcastList(ChainingBroadcastList[Tag]):
    parent: ChainingBroadcastList[Tag | None]
    previous_element: ChainingBroadcastList[PageElement | None]
    next_element: ChainingBroadcastList[PageElement | None]
    next_sibling: ChainingBroadcastList[PageElement | None]
    previous_sibling: ChainingBroadcastList[PageElement | None]
    def setup(
        self,
        parent: Tag | None = None,
        previous_element: PageElement | None = None,
        next_element: PageElement | None = None,
        previous_sibling: PageElement | None = None,
        next_sibling: PageElement | None = None,
    ) -> ChainingBroadcastList[None]: ...
    def format_string(self, s: str, formatter: Formatter | str | None) -> ChainingBroadcastList[str]: ...
    def formatter_for_name(self, formatter: Formatter | str | _EntitySubstitution) -> ChainingBroadcastList[None]: ...
    nextSibling: ChainingBroadcastList[PageElement | None]
    previousSibling: ChainingBroadcastList[PageElement | None]
    @property
    def stripped_strings(self) -> ChainingBroadcastList[Iterator[str]]: ...
    def get_text(self, separator: str = "", strip: bool = False, types: tuple[type[NavigableString], ...] = ...) -> ChainingBroadcastList[str]: ...
    getText = get_text
    @property
    def text(self) -> ChainingBroadcastList[str]: ...
    def replace_with(self, *args: PageElement | str) -> ChainingBroadcastList[Self]: ...
    replaceWith = replace_with
    def unwrap(self) -> ChainingBroadcastList[Self]: ...
    replace_with_children = unwrap
    replaceWithChildren = unwrap
    def wrap(self, wrap_inside: _PageElementT) -> ChainingBroadcastList[_PageElementT]: ...
    def extract(self, _self_index: int | None = None) -> ChainingBroadcastList[Self]: ...
    def Einsert(self, position: int, new_child: PageElement | str) -> ChainingBroadcastList[None]: ...
    def Eappend(self, tag: PageElement | str) -> ChainingBroadcastList[None]: ...
    def Eextend(self, tags: Iterable[PageElement | str]) -> ChainingBroadcastList[None]: ...
    def insert_before(self, *args: PageElement | str) -> ChainingBroadcastList[None]: ...
    def insert_after(self, *args: PageElement | str) -> ChainingBroadcastList[None]: ...
    def find_next(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    findNext = find_next
    def find_all_next(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[PageElement]]: ...
    findAllNext = find_all_next
    def find_next_sibling(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    findNextSibling = find_next_sibling
    def find_next_siblings(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[PageElement]]: ...
    findNextSiblings = find_next_siblings
    fetchNextSiblings = find_next_siblings
    def find_previous(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    findPrevious = find_previous
    def find_all_previous(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[PageElement]]: ...
    findAllPrevious = find_all_previous
    fetchPrevious = find_all_previous
    def find_previous_sibling(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    findPreviousSibling = find_previous_sibling
    def find_previous_siblings(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[PageElement]]: ...
    findPreviousSiblings = find_previous_siblings
    fetchPreviousSiblings = find_previous_siblings
    def find_parent(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | None]: ...
    findParent = find_parent
    def find_parents(
        self,
        name: _Strainable | SoupStrainer | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[Tag]]: ...
    findParents = find_parents
    fetchParents = find_parents
    @property
    def next(self) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    @property
    def previous(self) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    @property
    def next_elements(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    @property
    def next_siblings(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    @property
    def previous_elements(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    @property
    def previous_siblings(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    @property
    def parents(self) -> ChainingBroadcastList[Iterable[Tag]]: ...
    @property
    def decomposed(self) -> ChainingBroadcastList[bool]: ...
    def nextGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def nextSiblingGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def previousGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def previousSiblingGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def parentGenerator(self) -> ChainingBroadcastList[Iterable[Tag]]: ...

    # TAGS

    parser_class: ChainingBroadcastList[type[BeautifulSoup] | None]
    name: ChainingBroadcastList[str]
    namespace: ChainingBroadcastList[str | None]
    prefix: ChainingBroadcastList[str | None]
    sourceline: ChainingBroadcastList[int | None]
    sourcepos: ChainingBroadcastList[int | None]
    known_xml: ChainingBroadcastList[bool | None]
    attrs: ChainingBroadcastList[dict[str, str]]
    contents: ChainingBroadcastList[list[PageElement]]
    hidden: ChainingBroadcastList[bool]
    can_be_empty_element: ChainingBroadcastList[bool | None]
    cdata_list_attributes: ChainingBroadcastList[list[str] | None]
    preserve_whitespace_tags: ChainingBroadcastList[list[str] | None]
    # def __init__(
    #     self,
    #     parser: BeautifulSoup | None = None,
    #     builder: TreeBuilder | None = None,
    #     name: str | None = None,
    #     namespace: str | None = None,
    #     prefix: str | None = None,
    #     attrs: dict[str, str] | None = None,
    #     parent: Tag | None = None,
    #     previous: PageElement | None = None,
    #     is_xml: bool | None = None,
    #     sourceline: int | None = None,
    #     sourcepos: int | None = None,
    #     can_be_empty_element: bool | None = None,
    #     cdata_list_attributes: list[str] | None = None,
    #     preserve_whitespace_tags: list[str] | None = None,
    #     interesting_string_types: type[NavigableString] | tuple[type[NavigableString], ...] | None = None,
    #     namespaces: dict[str, str] | None = None,
    # ) -> ChainingBroadcastList[None]: ...
    parserClass: ChainingBroadcastList[type[BeautifulSoup] | None]
    def __copy__(self) -> ChainingBroadcastList[Self]: ...
    @property
    def is_empty_element(self) -> ChainingBroadcastList[bool]: ...
    @property
    def isSelfClosing(self) -> ChainingBroadcastList[bool]: ...
    @property
    def string(self) -> ChainingBroadcastList[str | None]: ...
    @string.setter
    def string(self, string: str) -> ChainingBroadcastList[None]: ...
    DEFAULT_INTERESTING_STRING_TYPES: ChainingBroadcastList[tuple[type[NavigableString], ...]]
    @property
    def strings(self) -> ChainingBroadcastList[Iterable[str]]: ...
    def decompose(self) -> ChainingBroadcastList[None]: ...
    def Eclear(self, decompose: bool = False) -> ChainingBroadcastList[None]: ...
    def smooth(self) -> ChainingBroadcastList[None]: ...
    def Eindex(self, element: PageElement) -> ChainingBroadcastList[int]: ...
    def get(self, key: str, default: str | list[str] | None = None) -> ChainingBroadcastList[str | list[str] | None]: ...
    def get_attribute_list(self, key: str, default: str | list[str] | None = None) -> ChainingBroadcastList[list[str]]: ...
    def has_attr(self, key: str) -> ChainingBroadcastList[bool]: ...
    def E__hash__(self) -> ChainingBroadcastList[int]: ...
    def E__getitem__(self, key: str) -> ChainingBroadcastList[str | list[str]]: ...
    def E__iter__(self) -> ChainingBroadcastList[Iterator[PageElement]]: ...
    def E__len__(self) -> ChainingBroadcastList[int]: ...
    def E__contains__(self, x: object) -> ChainingBroadcastList[bool]: ...
    def E__bool__(self) -> ChainingBroadcastList[bool]: ...
    def __setitem__(self, key: str, value: str | list[str]) -> ChainingBroadcastList[None]: ...
    def E__delitem__(self, key: str) -> ChainingBroadcastList[None]: ...
    def __getattr__(self, tag: str) -> ChainingBroadcastList[Tag | None]: ...
    def E__eq__(self, other: object) -> ChainingBroadcastList[bool]: ...
    def E__ne__(self, other: object) -> ChainingBroadcastList[bool]: ...
    def __unicode__(self) -> ChainingBroadcastList[str]: ...
    def encode(
        self,
        encoding: str = "utf-8",
        indent_level: int | None = None,
        formatter: str | Formatter = "minimal",
        errors: str = "xmlcharrefreplace",
    ) -> ChainingBroadcastList[bytes]: ...
    def decode(
        self,
        indent_level: int | None = None,
        eventual_encoding: str = "utf-8",
        formatter: str | Formatter = "minimal",
        iterator: Iterator[PageElement] | None = None,
    ) -> ChainingBroadcastList[str]: ...
    @overload
    def prettify(self, encoding: str, formatter: str | Formatter = "minimal") -> ChainingBroadcastList[bytes]: ...
    @overload
    def prettify(self, encoding: None = None, formatter: str | Formatter = "minimal") -> ChainingBroadcastList[str]: ...
    def decode_contents(
        self, indent_level: int | None = None, eventual_encoding: str = "utf-8", formatter: str | Formatter = "minimal"
    ) -> ChainingBroadcastList[str]: ...
    def encode_contents(
        self, indent_level: int | None = None, encoding: str = "utf-8", formatter: str | Formatter = "minimal"
    ) -> ChainingBroadcastList[bytes]: ...
    def renderContents(self, encoding: str = "utf-8", prettyPrint: bool = False, indentLevel: int = 0) -> ChainingBroadcastList[bytes]: ...
    def find(
        self,
        name: _Strainable | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        recursive: bool = True,
        string: _Strainable | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[Tag | NavigableString | None]: ...
    findChild = find
    def find_all(
        self,
        name: _Strainable | None = None,
        attrs: dict[str, _Strainable] | _Strainable = {},
        recursive: bool = True,
        string: _Strainable | None = None,
        limit: int | None = None,
        **kwargs: _Strainable,
    ) -> ChainingBroadcastList[ResultSet[Any]]: ...
    __call__ = find_all
    findAll = find_all
    findChildren = find_all
    @property
    def children(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    @property
    def descendants(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def select_one(
        self, selector: str, namespaces: Incomplete | None = None, *, flags: int = ..., custom: dict[str, str] | None = ...
    ) -> ChainingBroadcastList[Tag | None]: ...
    def select(
        self,
        selector: str,
        namespaces: Incomplete | None = None,
        limit: int | None = None,
        *,
        flags: int = ...,
        custom: dict[str, str] | None = ...,
    ) -> ChainingBroadcastList[ResultSet[Tag]]: ...
    def childGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def recursiveChildGenerator(self) -> ChainingBroadcastList[Iterable[PageElement]]: ...
    def has_key(self, key: str) -> ChainingBroadcastList[bool]: ...


from __future__ import annotations


class EmptyResultError(Exception):
    def __init__(self, error_message: str, selector: str | None = None, url=None, status_code: int | None = None) -> None:
        if not url:
            url = "(can't show URL probably because `markup_or_response` parameter was string)"

        error_message = error_message.rstrip()

        error_string = (
            f'{error_message} '
            'This error happens probably because of invalid selector or URL. '
            'Check if both selector and URL are valid. '
            'Set to False `no_empty_result` if empty list is intended. '
            'It may also because of selector is not matched with URL.\n'
            f'selector: {selector!r}, URL: {url}'
        )

        if str(status_code)[0] in {'4', '5'}:
            error_string = (f'WARNING: status code (HTTP{status_code}) looks odd. '
                            f'check your identity, headers, and URL again.\n{error_string}')

        super().__init__(error_string)


class NoParserError(Exception):
    pass


class NotAResponseError(Exception):
    pass



#### IMPORTS FOR TYPE HINT

# from _typeshed import Incomplete
# from collections.abc import Mapping
# from typing_extensions import TypeAlias

# from requests.models import Response
# from .sessions import RequestsCookieJar, _Auth, _Cert, _Data, _Files, _HooksInput, _Params, _TextMapping, _Timeout, _Verify

from __future__ import annotations

# from _typeshed import Incomplete, SupportsItems, SupportsRead, Unused
from _typeshed import Incomplete, SupportsItems, SupportsRead
from collections.abc import Callable, Iterable, Mapping, MutableMapping
from typing import Any
# from typing_extensions import Self, TypeAlias, TypedDict
from typing_extensions import TypeAlias

# from urllib3._collections import RecentlyUsedContainer

# from requests import adapters, auth as _auth, compat, cookies, exceptions, hooks, models, status_codes, utils
from requests import auth as _auth, cookies, models
from requests.models import Response
from requests.structures import CaseInsensitiveDict as CaseInsensitiveDict

from .response_proxy import ResponseProxy

# _BaseAdapter: TypeAlias = adapters.BaseAdapter
# OrderedDict = compat.OrderedDict
# cookiejar_from_dict = cookies.cookiejar_from_dict
# extract_cookies_to_jar = cookies.extract_cookies_to_jar
RequestsCookieJar = cookies.RequestsCookieJar
# merge_cookies = cookies.merge_cookies
# Request = models.Request
PreparedRequest = models.PreparedRequest
# DEFAULT_REDIRECT_LIMIT = models.DEFAULT_REDIRECT_LIMIT
# default_hooks = hooks.default_hooks
# dispatch_hook = hooks.dispatch_hook
# to_key_val_list = utils.to_key_val_list
# default_headers = utils.default_headers  # redefine
# to_native_string = utils.to_native_string
# TooManyRedirects = exceptions.TooManyRedirects
# InvalidSchema = exceptions.InvalidSchema
# ChunkedEncodingError = exceptions.ChunkedEncodingError
# ContentDecodingError = exceptions.ContentDecodingError
# HTTPAdapter = adapters.HTTPAdapter
# requote_uri = utils.requote_uri
# get_environ_proxies = utils.get_environ_proxies
# get_netrc_auth = utils.get_netrc_auth
# should_bypass_proxies = utils.should_bypass_proxies
# get_auth_from_url = utils.get_auth_from_url
# codes = status_codes.codes
# REDIRECT_STATI = models.REDIRECT_STATI
# 
# def merge_setting(request_setting, session_setting, dict_class=...): ...
# def merge_hooks(request_hooks, session_hooks, dict_class=...): ...

# class SessionRedirectMixin:
#     def resolve_redirects(
#         self,
#         resp,
#         req,
#         stream: bool = False,
#         timeout: Incomplete | None = None,
#         verify: bool = True,
#         cert: Incomplete | None = None,
#         proxies: Incomplete | None = None,
#         yield_requests: bool = False,
#         **adapter_kwargs,
#     ): ...
#     def rebuild_auth(self, prepared_request, response): ...
#     def rebuild_proxies(self, prepared_request, proxies): ...
#     def should_strip_auth(self, old_url, new_url): ...
#     def rebuild_method(self, prepared_request: PreparedRequest, response: Response) -> None: ...
#     def get_redirect_target(self, resp: Response) -> str | None: ...

_Data: TypeAlias = (
    # used in requests.models.PreparedRequest.prepare_body
    #
    # case: is_stream
    # see requests.adapters.HTTPAdapter.send
    # will be sent directly to http.HTTPConnection.send(...) (through urllib3)
    Iterable[bytes]
    # case: not is_stream
    # will be modified before being sent to urllib3.HTTPConnectionPool.urlopen(body=...)
    # see requests.models.RequestEncodingMixin._encode_params
    # see requests.models.RequestEncodingMixin._encode_files
    # note that keys&values are converted from Any to str by urllib.parse.urlencode
    | str
    | bytes
    | SupportsRead[str | bytes]
    | list[tuple[Any, Any]]
    | tuple[tuple[Any, Any], ...]
    | Mapping[Any, Any]
)
_Auth: TypeAlias = tuple[str, str] | _auth.AuthBase | Callable[[PreparedRequest], PreparedRequest]
_Cert: TypeAlias = str | tuple[str, str]
# Files is passed to requests.utils.to_key_val_list()
_FileName: TypeAlias = str | None
_FileContent: TypeAlias = SupportsRead[str | bytes] | str | bytes
_FileContentType: TypeAlias = str
_FileCustomHeaders: TypeAlias = Mapping[str, str]
_FileSpecTuple2: TypeAlias = tuple[_FileName, _FileContent]
_FileSpecTuple3: TypeAlias = tuple[_FileName, _FileContent, _FileContentType]
_FileSpecTuple4: TypeAlias = tuple[_FileName, _FileContent, _FileContentType, _FileCustomHeaders]
_FileSpec: TypeAlias = _FileContent | _FileSpecTuple2 | _FileSpecTuple3 | _FileSpecTuple4
_Files: TypeAlias = Mapping[str, _FileSpec] | Iterable[tuple[str, _FileSpec]]
_Hook: TypeAlias = Callable[[Response], Any]
_HooksInput: TypeAlias = Mapping[str, Iterable[_Hook] | _Hook]

_ParamsMappingKeyType: TypeAlias = str | bytes | int | float
_ParamsMappingValueType: TypeAlias = str | bytes | int | float | Iterable[str | bytes | int | float] | None
_Params: TypeAlias = (
    SupportsItems[_ParamsMappingKeyType, _ParamsMappingValueType]
    | tuple[_ParamsMappingKeyType, _ParamsMappingValueType]
    | Iterable[tuple[_ParamsMappingKeyType, _ParamsMappingValueType]]
    | str
    | bytes
)
_TextMapping: TypeAlias = MutableMapping[str, str]
# _HeadersUpdateMapping: TypeAlias = Mapping[str, str | bytes | None]
_Timeout: TypeAlias = float | tuple[float, float] | tuple[float, None]
_Verify: TypeAlias = bool | str

# class _Settings(TypedDict):
#     verify: _Verify | None
#     proxies: _TextMapping
#     stream: bool
#     cert: _Cert | None

_HeadersMapping: TypeAlias = Mapping[str, str | bytes]

DEFAULT_HEADERS: dict[str, str]

def request(
    method: str | bytes,
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def get(
    url: str | bytes,
    params: _Params | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def options(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def head(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def post(
    url: str | bytes,
    data: _Data | None = None,
    json: Incomplete | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
) -> ResponseProxy: ...
def put(
    url: str | bytes,
    data: _Data | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def patch(
    url: str | bytes,
    data: _Data | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def delete(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...

# CACHED REQESTS

def crequest(
    method: str | bytes,
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...

# CACHED REQUESTS

def cget(
    url: str | bytes,
    params: _Params | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def coptions(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def chead(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def cpost(
    url: str | bytes,
    data: _Data | None = None,
    json: Incomplete | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
) -> ResponseProxy: ...
def cput(
    url: str | bytes,
    data: _Data | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def cpatch(
    url: str | bytes,
    data: _Data | None = None,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
def cdelete(
    url: str | bytes,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...

# ASYNCRONIZED, CACHED REQESTS

async def acrequest(
    method: str | bytes,
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def acget(
    url: str | bytes,
    params: _Params | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def acoptions(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def achead(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def acpost(
    url: str | bytes,
    data: _Data | None = None,
    json: Incomplete | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
) -> ResponseProxy: ...
async def acput(
    url: str | bytes,
    data: _Data | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def acpatch(
    url: str | bytes,
    data: _Data | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def acdelete(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...

# ASYNCRONIZED REQESTS

async def arequest(
    method: str | bytes,
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def aget(
    url: str | bytes,
    params: _Params | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def aoptions(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def ahead(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def apost(
    url: str | bytes,
    data: _Data | None = None,
    json: Incomplete | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
) -> ResponseProxy: ...
async def aput(
    url: str | bytes,
    data: _Data | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def apatch(
    url: str | bytes,
    data: _Data | None = None,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...
async def adelete(
    url: str | bytes,
    run_in_executor: bool = True,
    *,
    attempts: int | None = ...,
    raise_for_status: bool | None = ...,
    avoid_sslerror: bool | None = ...,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: RequestsCookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: Incomplete | None = ...,
) -> ResponseProxy: ...


"""avoid SSLError due to UNSAFE_LEGACY_RENEGOTIATION_DISABLED

from https://stackoverflow.com/a/71646353/21997874
"""

import ssl
import urllib3
import requests.adapters


class CustomHttpAdapter(requests.adapters.HTTPAdapter):
    """Transport adapter that allows us to use custom ssl_context."""
    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, ssl_context=self.ssl_context)


ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
ctx.options |= 0x4
adapter = CustomHttpAdapter(ctx)


def make_session_sslerror_free(session: requests.Session, prefix: str = 'https://') -> None:
    session.mount(prefix, adapter)



from __future__ import annotations

# '#': 작동하긴 하지만 공식적으로 지원하지 않는 모듈
# '# #': 작동하지 않는 모듈
from requests import (


    ConnectTimeout,
    ConnectionError,
    # DependencyWarning,
    FileModeWarning,
    HTTPError,
    JSONDecodeError,
    # NullHandler,
    PreparedRequest,
    ReadTimeout,
    Request,
    RequestException,
    # RequestsDependencyWarning,
    Response,
    # # Session,
    Timeout,
    TooManyRedirects,
    URLRequired,
    __author__,
    __author_email__,
    __build__,
    # # __builtins__,
    # # __cached__,
    __cake__,
    __copyright__,
    __description__,
    # # __doc__,
    # # __file__,
    __license__,
    # # __loader__,
    # # __name__,
    # # __package__,
    # # __path__,
    # # __spec__,
    __title__,
    __url__,
    __version__,
    # _check_cryptography,
    # _internal_utils,
    adapters,
    # # api,
    auth,
    certs,
    # chardet_version,
    # charset_normalizer_version,
    check_compatibility,
    codes,
    compat,
    cookies,
    # # delete,
    exceptions,
    # # get,
    # # head,
    hooks,
    # logging,
    models,
    # # options,
    packages,
    # # patch,
    # # post,
    # # put,
    # # request,
    session,
    sessions,
    # # ssl,
    status_codes,
    structures,
    # urllib3,
    utils,
    # warnings,
)

from .api_with_tools import (
    request,
    get, options, head, post, put, patch, delete,
    cget, coptions, chead, cpost, cput, cpatch, cdelete,
    acget, acoptions, achead, acpost, acput, acpatch, acdelete,
    aget, aoptions, ahead, apost, aput, apatch, adelete,
)
from .sessions_with_tools import Session



"""Replacement of BeatifulSoup."""

from __future__ import annotations
import contextlib

from typing import overload, Literal
from typing_extensions import Self
from requests.models import Response
from bs4 import BeautifulSoup, FeatureNotFound
from bs4.element import Tag, ResultSet

from .exceptions import (
    NoParserError,
    EmptyResultError,
)
from .broadcast_list import TagBroadcastList

Parsers = Literal["html.parser", "html", "lxml", "lxml-xml", "xml", "html5lib", "html5"]


class SoupTools:
    def __init__(self, text: str) -> None:
        self.text: str = text
        self.response: Response | None = None

    @classmethod
    def from_response(cls, response: Response) -> Self:
        new = cls(response.text)
        new.response = response
        return new

    def __getattr__(self, name: str):
        if self.response is not None:
            with contextlib.suppress(AttributeError):
                # 만약 return 과정에서 AttributeError가 나면
                # pass되고 아래에 있는 AttributeError가 사용됨.
                return getattr(self.response, name)

        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def soup(
        self,
        parser: Parsers | None = None,
    ) -> BeautifulSoup:
        if parser is None:
            # 없으면 warning이 뜸
            parser = 'html.parser'
        return BeautifulSoup(self.text, parser)

    @overload
    def soup_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
        use_broadcast_list: Literal[True] = ...,
    ) -> TagBroadcastList:
        ...

    @overload
    def soup_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
        use_broadcast_list: Literal[False] = ...,
    ) -> ResultSet[Tag]:
        ...

    @overload
    def soup_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
        use_broadcast_list: bool = True,
    ) -> ResultSet[Tag] | TagBroadcastList:
        ...

    def soup_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
        use_broadcast_list: bool = True,
    ) -> ResultSet[Tag] | TagBroadcastList:
        """response.soup(parser, **kwargs).select(selector)와 거의 같습니다만 no_empty_result라는 강력한 추가 기능을 제공합니다.

        Args:
            self (str | Response): markup or response you want to parse.
            selector (str): BeatifulSoup.select의 selector입니다.
            no_empty_result (bool, optional): 결과가 빈 리스트라면 EmptyResultError를 냅니다. Defaults to False.
            parser (Parsers, optional): BeatifulSoup의 parser입니다. Defaults to 'html.parser'.

        Raises:
            EmptyResultError: 결과가 빈 리스트이고 no_empty_result가 참이라면 EmptyResultError를 냅니다.
                결과가 None일때 오류를 내는 것이 아니라는 점을 주의하세요.

        Returns:
            ResultSet[Tag]
        """
        selected = self.soup(parser).select(selector)
        if not no_empty_result or selected != []:
            return TagBroadcastList(selected) if use_broadcast_list else selected

        if self.response is None:
            raise EmptyResultError(
                'Result of select is empty list("[]").',
                selector=selector,
            )

        raise EmptyResultError(
            'Result of select is empty list("[]").',
            selector=selector,
            url=self.url,
            status_code=self.status_code,
        )

    @overload
    def soup_select_one(
        self,
        selector: str,
        no_empty_result: Literal[False] = ...,
        parser: Parsers | None = None,
    ) -> Tag | None:
        ...

    @overload
    def soup_select_one(
        self,
        selector: str,
        no_empty_result: Literal[True] = ...,
        parser: Parsers | None = None,
    ) -> Tag:
        ...

    @overload
    def soup_select_one(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
    ) -> Tag | None:
        ...

    def soup_select_one(
        self,
        selector: str,
        no_empty_result: bool = False,
        parser: Parsers | None = None,
    ) -> Tag | None:
        """response.soup(parser, **kwargs).select_one(selector)와 거의 같습니다만 no_empty_result라는 강력한 추가 기능을 제공합니다.

        Args:
            self (str | Response): markup or response you want to parse.
                (ResponesProxy에서 사용할 경우 없음)
            selector (str): BeatifulSoup.select_one의 selector입니다.
            no_empty_result (bool, optional): 결과(리턴값)가 None라면 EmptyResultError를 냅니다.
                typing과 오류 제거에 상당한 도움을 줍니다. 기존의 BeatifulSoup.select_one의 경우에는
                결과값이 None이거나 Tag였습니다. 따라서 BeatifulSoup.select_one(selector).text와 같은
                코드를 짤 때 정적 타입 검사기에서 오류를 내기 일쑤였고, 실제로 해당 코드 실행 결과가 None일 경우
                오류가 났습니다. no_empty_list를 이용해 불명확한 오류 대신 EmptyResultError를 내보내고
                타입 검사기의 오류도 피할 수 있어 좋은 기능입니다. 하지만 어떠한 이유로든지 이 기능을 사용하고
                싶지 않다면 간단히 그냥 값을 False로 하면 됩니다. Defaults to True.
            parser (Parsers, optional): BeatifulSoup의 parser입니다. Defaults to 'html.parser'.

        Raises:
            EmptyResultError: 결과가 None이고 no_empty_result가 참이라면 EmptyResultError를 냅니다.

        Returns:
            Tag | None: no_empty_result가 False일 경우(기본값)
            Tag: no_empty_result가 True일 경우(정적 검사기에 반영됨)
        """
        select_results = self.soup(parser).select_one(selector)
        if not no_empty_result or select_results is not None:
            return select_results

        if self.response is None:
            raise EmptyResultError(
                'Result of select_one is None.',
                selector=selector,
            )

        raise EmptyResultError(
            'Result of select_one is None.',
            selector=selector,
            url=self.url,
            status_code=self.status_code
        )

    # XML

    def xml(self) -> BeautifulSoup:
        """parser가 xml인 .soup()입니다. 자세한 내용은 .soup()의 docstring을 확인하세요."""
        # functools.partial을 사용할까도 했지만 그러면 type hint와 docstring 사용이 어렵다.
        return self.soup(parser='xml')

    @overload
    def xml_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        use_broadcast_list: Literal[True] = ...,
    ) -> TagBroadcastList:
        ...

    @overload
    def xml_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        use_broadcast_list: Literal[False] = ...,
    ) -> ResultSet[Tag]:
        ...

    @overload
    def xml_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        use_broadcast_list: bool = True,
    ) -> ResultSet[Tag] | TagBroadcastList:
        ...

    def xml_select(
        self,
        selector: str,
        no_empty_result: bool = False,
        use_broadcast_list: bool = True,
    ) -> ResultSet[Tag] | TagBroadcastList:
        """parser가 xml인 .soup_select()입니다. 자세한 내용은 .soup_select()의 docstring을 확인하세요."""
        return self.soup_select(selector, no_empty_result, 'xml', use_broadcast_list)

    @overload
    def xml_select_one(
        self,
        selector: str,
        no_empty_result: Literal[True] = ...,
    ) -> Tag:
        ...

    @overload
    def xml_select_one(
        self,
        selector: str,
        no_empty_result: Literal[True] = ...,
    ) -> Tag:
        ...

    @overload
    def xml_select_one(
        self,
        selector: str,
        no_empty_result: bool = False,
    ) -> Tag | None:
        ...

    def xml_select_one(
        self,
        selector: str,
        no_empty_result: bool = False,
    ) -> Tag | None:
        """parser가 xml인 .soup_select_one()입니다. 자세한 내용은 .soup_select_one()의 docstring을 확인하세요."""
        return self.soup_select_one(selector, no_empty_result, 'xml')



import functools

from . import api_with_tools


class CustomDefaults:
    def __init__(
        self,
        **kwargs
    ) -> None:
        self.defaults = kwargs

    def __getattr__(self, name):
        return functools.partial(getattr(api_with_tools, name), **self.defaults)



# TODO: raise_for_status 추가
"""
requests_plus.requests_api_with_more_tools
~~~~~~~~~~~~

This module implements the Requests API and more tools.

:copyright: (c) 2023 by ilotoki0804.
:license: MIT License, see LICENSE for more details.
"""

from __future__ import annotations

from functools import lru_cache
import asyncio
import logging
# from typing import Literal, Any, TypedDict
from urllib import parse

from requests import sessions, exceptions

from .response_proxy import ResponseProxy
from .dealing_unhashable_args import freeze_dict_and_list
from .contants import DEFAULT_HEADERS
from .avoid_sslerror import make_session_sslerror_free

__all__ = (
    'request',
    'get', 'options', 'head', 'post', 'put', 'patch', 'delete',
    'cget', 'coptions', 'chead', 'cpost', 'cput', 'cpatch', 'cdelete',
    'acget', 'acoptions', 'achead', 'acpost', 'acput', 'acpatch', 'acdelete',
    'aget', 'aoptions', 'ahead', 'apost', 'aput', 'apatch', 'adelete',
)


def request(
    method,
    url,
    attempts: int | None = None,
    avoid_sslerror: bool | None = None,
    raise_for_status: bool | None = None,
    **kwargs
):
    """기본값, 재시도 횟수 등 추가 기능이 들어간 requests_plus 버전의 requests.request 구현입니다.

    ## 추가된 기능
    만약 명시하지 않았다면 기본값이 적용됩니다. timeout의 기본값은 120, headers의 기본값은 간단한 user agent, attempts의 기본값은 1,
    avoid_sslerror은 False, raise_for_status는 False입니다.

    ## 기존 requests 라이브러리에는 없는 parameter
    :param attempts: (optional) request가 ConnectionError를 받았을 때 같은 요청을 몇 번 다시 실행할 것인지를 정합니다.
    :type attempts: positive integer

    ## 기존 requests 라이브러리에 있었던 parameter
    :param method: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
    :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
        ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
        or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
        defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
        to add for the file.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How many seconds to wait for the server to send data
        before giving up, as a float, or a :ref:`(connect timeout, read
        timeout) <timeouts>` tuple.
    :type timeout: float or tuple
    :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
    :type allow_redirects: bool
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response

    Usage::

      >>> from resoup import requests
      >>> req = requests.request('GET', 'https://httpbin.org/get')
      >>> req
      <Response [200]>
    """

    kwargs.setdefault('headers', DEFAULT_HEADERS)
    kwargs.setdefault('timeout', 120)
    raise_for_status = False if raise_for_status is None else raise_for_status
    attempts = 1 if attempts is None else attempts
    avoid_sslerror = False if avoid_sslerror is None else avoid_sslerror

    last_exception = None
    for _ in range(attempts):
        try:
            with sessions.Session() as session:
                if avoid_sslerror:
                    prefix = parse.urlparse(url).scheme + '://'
                    make_session_sslerror_free(session, prefix=prefix)
                response = ResponseProxy(session.request(method=method, url=url, **kwargs))
                if raise_for_status:
                    response.raise_for_status()
        except exceptions.ConnectionError as e:
            if attempts == 1:
                raise
            logging.warning('Retring...')
            last_exception = e
        else:
            if last_exception is not None:
                logging.warning(f'Sucessfully retried from {url}')
            return response

    raise ConnectionError(f'Trying {attempts} times but failed to get data.\nURL: {url}') from last_exception


def get(url, params=None, **kwargs):
    r"""Sends a GET request.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("get", url, params=params, **kwargs)


def options(url, **kwargs):
    r"""Sends an OPTIONS request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("options", url, **kwargs)


def head(url, **kwargs):
    r"""Sends a HEAD request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes. If
        `allow_redirects` is not provided, it will be set to `False` (as
        opposed to the default :meth:`request` behavior).
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault("allow_redirects", False)
    return request("head", url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    r"""Sends a POST request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("post", url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    r"""Sends a PUT request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("put", url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    r"""Sends a PATCH request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("patch", url, data=data, **kwargs)


def delete(url, **kwargs):
    r"""Sends a DELETE request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request("delete", url, **kwargs)


# CACHED REQESTS


@freeze_dict_and_list()
@lru_cache
def cget(url, params=None, **kwargs):
    "cached requests.get"
    return get(url, params=params, **kwargs)


@freeze_dict_and_list()
@lru_cache
def coptions(url, **kwargs):
    "cached requests.options"
    return options(url, **kwargs)


@freeze_dict_and_list()
@lru_cache
def chead(url, **kwargs):
    "cached requests.head"
    kwargs.setdefault("allow_redirects", False)
    return head(url, **kwargs)


@freeze_dict_and_list()
@lru_cache
def cpost(url, data=None, json=None, **kwargs):
    "cached requests.post"
    return post(url, data=data, json=json, **kwargs)


@freeze_dict_and_list()
@lru_cache
def cput(url, data=None, **kwargs):
    "cached requests.put"
    return put(url, data=data, **kwargs)


@freeze_dict_and_list()
@lru_cache
def cpatch(url, data=None, **kwargs):
    "cached requests.patch"
    return patch(url, data=data, **kwargs)


@freeze_dict_and_list()
@lru_cache
def cdelete(url, **kwargs):
    "cached requests.delete"
    return delete(url, **kwargs)


# ASYNCRONIZED, CACHED REQESTS


async def acget(url, params=None, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.get"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: cget(url, **kwargs))
    return cget(url, params=params, **kwargs)


async def acoptions(url, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.options"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: coptions(url, **kwargs))
    return coptions(url, **kwargs)


async def achead(url, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.head"
    kwargs.setdefault("allow_redirects", False)
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: chead(url, **kwargs))
    return chead(url, **kwargs)


async def acpost(url, data=None, json=None, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.post"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: cpost(url, data=data, json=json, **kwargs))
    return cpost(url, data=data, json=json, **kwargs)


async def acput(url, data=None, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.put"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: cput(url, data=data, **kwargs))
    return cput(url, data=data, **kwargs)


async def acpatch(url, data=None, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.patch"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: cpatch(url, data=data, **kwargs))
    return cpatch(url, data=data, **kwargs)


async def acdelete(url, run_in_executor: bool = True, **kwargs):
    "asyncronized, cached requests.delete"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: cdelete(url, **kwargs))
    return cdelete(url, **kwargs)


# ASYNCRONIZED REQESTS


async def aget(url, params=None, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.get"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: get(url, **kwargs))
    return get(url, params=params, **kwargs)


async def aoptions(url, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.options"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: options(url, **kwargs))
    return options(url, **kwargs)


async def ahead(url, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.head"
    kwargs.setdefault("allow_redirects", False)
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: head(url, **kwargs))
    return head(url, **kwargs)


async def apost(url, data=None, json=None, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.post"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: post(url, data=data, json=json, **kwargs))
    return post(url, data=data, json=json, **kwargs)


async def aput(url, data=None, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.put"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: put(url, data=data, **kwargs))
    return put(url, data=data, **kwargs)


async def apatch(url, data=None, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.patch"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: patch(url, data=data, **kwargs))
    return patch(url, data=data, **kwargs)


async def adelete(url, run_in_executor: bool = True, **kwargs):
    "asyncronized requests.delete"
    if run_in_executor:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: delete(url, **kwargs))
    return delete(url, **kwargs)



def clean_headers(raw_headers: str):
    is_name = True
    name: str = ''
    headers = {}
    for i, line in enumerate(filter(None, raw_headers.splitlines())):
        if not is_name:
            headers[name] = line
            is_name = True
            continue

        if line[-1] != ':':
            raise ValueError(f'Unexpected string: {line} on {i + 1}th line.')

        name = line.removesuffix(':')
        is_name = False

    return headers



"""
██████╗░███████╗░██████╗░█████╗░██╗░░░██╗██████╗░
██╔══██╗██╔════╝██╔════╝██╔══██╗██║░░░██║██╔══██╗
██████╔╝█████╗░░╚█████╗░██║░░██║██║░░░██║██████╔╝
██╔══██╗██╔══╝░░░╚═══██╗██║░░██║██║░░░██║██╔═══╝░
██║░░██║███████╗██████╔╝╚█████╔╝╚██████╔╝██║░░░░░
╚═╝░░╚═╝╚══════╝╚═════╝░░╚════╝░░╚═════╝░╚═╝░░░░░

REquests + beautifulSOUP

Various convenient features related to requests.
"""

__title__ = "resoup"
__description__ = 'Various convenient features related to requests and BeautifulSoup.'
__url__ = "https://github.com/ilotoki0804/resoup"
__raw_source_url__ = "https://raw.githubusercontent.com/ilotoki0804/resoup/master"
__version_info__ = (0, 5, 0)
__version__ = str.join('.', map(str, __version_info__))
__author__ = "ilotoki0804"
__author_email__ = "ilotoki0804@gmail.com"
__license__ = "MIT License"

from . import requests_ as requests
from .custom_defaults import CustomDefaults
from .header_utils import clean_headers
from .sessions_with_tools import Session
from .souptools import SoupTools



from __future__ import annotations

from requests.models import Response

from .souptoolsclass import SoupTools


class ResponseProxy(Response, SoupTools):
    def __init__(self, response):
        state = response.__reduce__()[2]
        self.__setstate__(state)  # type: ignore
        self.response = response
